`timescale 1ns / 1ps
`default_nettype none

//******************************************************************************
//* Folyadkszint jelz perifria.                                             *
//*                                                                            *
//*  ------------     ----                                                     *
//* |          O-|-->|    |                                                    *
//* |          O-|-->|    | folyadkszint                                      *
//* |          O-|-->|    |     (0-8)       ------------                       *
//* |~~~~~~~~~~O-|-->|    |/--------------\| MicroBlaze |                      *
//* |~~~~~~~~~~O-|-->|    |\--------------/| CPU        |                      *
//* |~~~~~~~~~~O-|-->|    |    AXI4-Lite    ------------                       *
//* |~~~~~~~~~~O-|-->|    |                                                    *
//* |~~~~~~~~~~O-|-->|    |                                                    *
//*  ------------     ----                                                     *
//*    tartly      perifria                                                  *
//******************************************************************************
module fluid_level_indicator(
    //rajel s reset.
    input  wire        clk,         //100 MHz rendszerrajel
    input  wire        rst,         //Aktv magas szinkron reset
    
    //Regiszter rsi interfsz.
    input  wire [3:0]  wr_addr,     //rsi cm
    input  wire        wr_en,       //rs engedlyez jel
    input  wire [31:0] wr_data,     //rsi adat
    input  wire [3:0]  wr_strb,     //Bjt engedlyez jelek
    
    //Regiszter olvassi interfsz.
    input  wire [3:0]  rd_addr,     //Olvassi cm
    input  wire        rd_en,       //Olvass engedlyez jel
    output reg  [31:0] rd_data,     //Olvassi adat      
    
    //Adat a folyadk rzkelktl.
    input  wire [7:0]  sensor_in,
    
    //Megszaktskr kimenet.
    output wire        irq
);

//******************************************************************************
//* A szenzor bemeneteket 10 Hz frekvencival mintavtelezzk. Az temez jel  *
//* egy 24 bites szmllval llthat el (9999999 - 0 => 24 bit).            *
//******************************************************************************
reg  [23:0] clk_div;
wire        clk_div_tc = (clk_div == 24'd0);

always @(posedge clk)
begin
   if (rst || clk_div_tc)
      clk_div <= 24'd9999999;
   else
      clk_div <= clk_div - 24'd1;
end

reg [7:0] sensor_reg;

always @(posedge clk)
begin
   if (rst)
      sensor_reg <= 8'd0;
   else
      if (clk_div_tc)
         sensor_reg <= sensor_in;
end

//******************************************************************************
//* A szenzor bemenet alapjn a folyadkszintet egy priorits enkderrel       *
//* llthatjuk el. A legnagyobb sorszm aktv bit hatrozza meg a           *
//* folyadkszint rtkt.                                                     *
//******************************************************************************
reg [3:0] fluid_level;

always @(posedge clk)
begin
   if (rst)
      fluid_level <= 4'd0;
   else
      casex (sensor_reg)
         8'b0000_0000: fluid_level <= 4'd0;
         8'b0000_0001: fluid_level <= 4'd1;
         8'b0000_001x: fluid_level <= 4'd2;
         8'b0000_01xx: fluid_level <= 4'd3;
         8'b0000_1xxx: fluid_level <= 4'd4;
         8'b0001_xxxx: fluid_level <= 4'd5;
         8'b001x_xxxx: fluid_level <= 4'd6;
         8'b01xx_xxxx: fluid_level <= 4'd7;
         8'b1xxx_xxxx: fluid_level <= 4'd8;
      endcase
end

//******************************************************************************
//* A hibajelzs ellltsa. rvnyes a szenzor bemeneten lv adat, ha       *
//* a legnagyobb sorszm aktv bemeneti bit alatti sszes bit is aktv.       *
//******************************************************************************
reg error;

always @(posedge clk)
begin
   if (rst)
      error <= 1'b0;
   else
      case (sensor_reg)
         8'b0000_0000: error <= 1'b0;
         8'b0000_0001: error <= 1'b0;
         8'b0000_0011: error <= 1'b0;
         8'b0000_0111: error <= 1'b0;
         8'b0000_1111: error <= 1'b0;
         8'b0001_1111: error <= 1'b0;
         8'b0011_1111: error <= 1'b0;
         8'b0111_1111: error <= 1'b0;
         8'b1111_1111: error <= 1'b0;
         default     : error <= 1'b1;
      endcase
end

//******************************************************************************
//* A regiszterek rsi s olvassi engedlyez jeleinek ellltsa.          *
//*                                                                            *
//* Folyadkszint regiszter     : BASEADDR+0x00, 32 bites, RD                  *
//* Megszakts engedlyez reg.: BASEADDR+0x04, 32 bites, R/W                 *
//* Megszakts flag regiszter  : BASEADDR+0x08, 32 bites, R/W1C               *
//******************************************************************************
//A folyadkszint regiszter olvass engedlyez jele.
wire lvl_rd = rd_en & (rd_addr[3:2] == 2'd0);

//A megszakts engedlyez regiszter rs s olvass engedlyez jele.
wire ier_wr = wr_en & (wr_addr[3:2] == 2'd1) & (wr_strb == 4'b1111);
wire ier_rd = rd_en & (rd_addr[3:2] == 2'd1);

//A megszakts flag regiszter rs s olvass engedlyez jele.
wire ifr_wr = wr_en & (wr_addr[3:2] == 2'd2) & (wr_strb == 4'b1111);
wire ifr_rd = rd_en & (rd_addr[3:2] == 2'd2);

//******************************************************************************
//* Folyadkszint regiszter: BASEADDR+0x00, 32 bites, csak olvashat           *
//*                                                                            *
//*    31    30          4     3     2     1     0                             *
//*  -----------------------------------------------                           *
//* |ERROR|  0   ....    0  |  folyadkszint (0-8)  |                          *
//*  -----------------------------------------------                           * 
//******************************************************************************
wire [31:0] lvl;

assign lvl[3:0]  = fluid_level;
assign lvl[30:4] = 27'd0;
assign lvl[31]   = error;

//******************************************************************************
//* Megszakts engedlyez reg.: BASEADDR+0x04, 32 bites, rhat/olvashat    *
//*                                                                            *
//*    31          5     4     3     2     1     0                             *
//*  -----------------------------------------------                           *
//* |  x    ....   x     x    x   |ERROR|EMPTY| FULL|                          *
//*  -----------------------------------------------                           *
//******************************************************************************
reg [2:0] ier;

always @(posedge clk)
begin
   if (rst)
      ier <= 3'b000;
   else
      if (ier_wr)
         ier <= wr_data[2:0];
end

//******************************************************************************
//* Megszakts flag regiszter: BASEADDR+0x08, 32 bites, olvashat s a jelzs *
//*                             '1' bersval trlhet                       *
//*                                                                            *
//*    31          5     4     3     2     1     0                             *
//*  -----------------------------------------------                           *
//* |  x    ....   x     x    x   |ERROR|EMPTY| FULL|                          *
//*  -----------------------------------------------                           *
//******************************************************************************
//Mintavtelezs a felfut l detektlshoz.
reg [1:0] lvl8_samples;
reg [1:0] lvl0_samples;
reg [1:0] err_samples;

always @(posedge clk)
begin
   if (rst)
   begin
      lvl8_samples <= 2'b11;
      lvl0_samples <= 2'b11;
      err_samples  <= 2'b11;
   end
   else
   begin
      lvl8_samples <= {lvl8_samples[0], (fluid_level == 4'd8)};
      lvl0_samples <= {lvl0_samples[0], (fluid_level == 4'd0)};
      err_samples  <= {err_samples[0], error};
   end
end

reg  [2:0] ifr;
wire [2:0] ifr_set;

//A tartly ppen megtelt (FULL): a folyadkszint 8-ra vlt -> felfut l detektls.
assign ifr_set[0] = (lvl8_samples == 2'b01);
//A tartly ppen kirlt (EMPTY): a folyadkszint 0-ra vlt -> felfut l detektls.
assign ifr_set[1] = (lvl0_samples == 2'b01);
//Hiba trtnt (ERROR): felfut l a hibajelzsen.
assign ifr_set[2] = (err_samples == 2'b01);

integer i;

//A megszakts flag regisztert egyetlen always blokkban rjuk le, FOR
//ciklussal indexelve a biteket. A bitek belltsa nagyobb priorits
//az '1' bersnak hatsra megtrtn trlsnl.
always @(posedge clk)
begin
   for (i = 0; i < 3; i = i + 1)
      if (rst)
         ifr[i] <= 1'b0;
      else
         if (ifr_set[i])
            ifr[i] <= 1'b1;
         else
            if (ifr_wr && wr_data[i])
               ifr[i] <= 1'b0;
end

//Jelezzk a megszaktskrst, ha van aktv esemny, amely engedlyezett is.
assign irq = |(ier & ifr);

//******************************************************************************
//* Az olvassi adatbusz meghajtsa.                                           *
//******************************************************************************
always @(*)
begin
	case ({ifr_rd, ier_rd, lvl_rd})
		3'b001 : rd_data <= lvl;
		3'b010 : rd_data <= {29'd0, ier};
		3'b100 : rd_data <= {29'd0, ifr};
		default: rd_data <= 32'd0;
	endcase
end

endmodule

`default_nettype wire
